// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef GESTALT_H
#define GESTALT_H

#include <iostream>
#include <string>
#include <sstream>

#include "types.h"
#include "vectormath.h"
#include "geomshape.h"
#include "estatshape.h"
#include "vdwshape.h"
#include "hbondshape.h"
#include "bondshape.h"
#include "hishape.h"
#include "externalshape.h"
#include "markershape.h"

namespace brown{
	
	//! @author jni,tg
	
	class GeomShape;
	class EstatShape;
	class vdWShape;
	class bondShape;
	class HiShape;
	class ExternalShape;
	class Matrix;

	//! Container of specialized shape instances. Sums up forces delivered back from those.
	class Gestalt {

		// public stuff
		public: 
		
		//! Constructor for an empty Gestalt with a label
		Gestalt(std::string lbl);
		//! Copy Constructor
		Gestalt(const Gestalt &);
		//! Destructor
		~Gestalt();
		
		//! Some diagnostic self-description
		std::string describe();
		
		//! Who am I?
		std::string getLabel() const { return myLabel; }
		
		//! Dump internal structure
		void dumpStructure();
		
		//! Do output: triggered from cup via molecule
		void positionToSnapshot(FILE *fp, const char *eiDi, const long int id_number);

		//! Reset the total force onto a molecule
		//! Called after domove(), before main loop, and after creating a new molecule
		void resetTotalForce();
		
		
		//! @name Handlers for GeomShape
		//! @{
		//
		//! Set pointer to existing GeomShape, for wall also pass on wall position
		void setGeomShape(GeomShape *gs) {
			myGeomShape = gs;
			myGeomShape->setPositionNormal(myWallsPosition, myWallsNormal);
		}
		//! Return pointer to my GeomShape
		GeomShape *getGeomShape() const{
			return(myGeomShape);
		}
		//! @}
		
		
		//! @name Handlers for EstatShape
		//! @{
		//
		//! Set Pointer to existing EstatShape, for wall also hand on wall position
		void setEstatShape(EstatShape *es) {
			myEstatShape = es;
			myEstatShape->setPositionNormal(myWallsPosition, myWallsNormal);
		}
		//! Return pointer to my EstatShape
		EstatShape *getEstatShape() const{
			return(myEstatShape);
		}
		//! @}
		
	
		//! @name Handlers for vdWShape
		//! @{
		//
		//! Set pointer to existing vdWShape, for wall also hand on wall position
		void setvdWShape(vdWShape *vs) {
			myvdWShape = vs;
			myvdWShape->setPositionNormal(myWallsPosition, myWallsNormal);
		}
		//! Return Pointer to my vdWShape
		vdWShape *getvdWShape() const {
			return(myvdWShape);
		}
		//! @}
		
		
		//! @name Handlers for HBondShape
		//! @{
		//
		//! Set pointer to existing HBondShape
		void setHBondShape(HBondShape *hs) {
			myhbondShape = hs;
		}
		//! Return pointer to my HBondShape
		HBondShape *getHBondShape() const {
			return(myhbondShape);
		}
		//! @}
		
		
		//! @name Handlers for BondShape
		//! @{
		//
		//! Set pointer to existing bondShape
		void setBondShape(BondShape *bs) {
			mybondShape = bs;
		}
		//! Return pointer to my bondShape
		BondShape *getBondShape() const {
			return(mybondShape);
		}
		//! @}

		
		//! @name Handlers for HiShape
		//! @{
		//
		//! Set pointer to existing HiShape
		void setHiShape(HiShape *hi) {
		    myHiShape = hi;
		}
		//! Return pointer to my HiShape
		HiShape *getHiShape() const {
		    return(myHiShape);
		}
		//! @}

		
		//! @name Handlers for ExternalShape
		//! @{
		//
		//! Set Pointer to existing ExternalShape
		void setExternalShape(ExternalShape *es) {
			myExternalShape = es;
		}
		//! Return pointer to my ExternalShape
		ExternalShape *getExternalShape() const {
			return(myExternalShape);
		}
		//! @}
		
		
		//! @name Handlers for MarkerShape
		//! @{
		//
		//! Set Pointer to existing MarkerShape
		void setMarkerShape(MarkerShape *ms) {
			myMarkerShape = ms;
		}
		//! Return pointer to my MarkerShape
		MarkerShape *getMarkerShape() const {
			return(myMarkerShape);
		}
		//! @}
		
		
		//! propagates the wall's normal down into the shapes -- if shapes already defined.. 
		//! Otherwise set*Shape has to do this job.
		void setPositionNormal(position *p, direction *n);
		
		
		//! return the pointers to the wall's position and normal
		position *getWallPosition() {
			return(myWallsPosition);
		}
		//! for testing purposes
		direction *getWallNormal() {
			return(myWallsNormal);
		}
		
		
		//! Return my current full position
		fullPosition *getMyPosition() {
			return(&myActualPosition);
		}
		
		//! Set my new position. This does not trigger a recalculation for the included shapes.
		void setMyPosition(const fullPosition &fp) {
			doublevec_set(myPosition.pos, fp.pos);
			doublevec_set(myActualPosition.pos, fp.pos);
			rotation_set(myPosition.rot, fp.rot);
			rotation_set(myActualPosition.rot, fp.rot);
		}
		
		//! teleport the gestalt and all its internal shapes to the new absolute fullPosition fp
		void teleport(const fullPosition &fp);
		
		//! displace gestalt and its internals by the increment dfp
		//! @{
		void move(const position &dx, const rotation &dphi);
		void move(const fullPosition &dfp);
		//! @}
		
		
		//! Re-calculate the actual positions of all attached shapes from the gestalt's actual position.
		//! Called when the Gestalt has been moved.
		void recalculate(const fullPosition &fp);

		
		bool isColliding(const Gestalt *other);
		
		//! @name Determine forces and effective displacements
		//! @{
		//
		//! Get a new set of random forces for the next timestep.
		//! Design decision to put this into its own function: allows for multi-step integrators
		void initRandomForces();
		
		//! Negotiate forces between the Gestalt objects, internally handled by the respective shapes.
		//! Set checkBonds to true for Gestalt objects on the same molecule, otherwise false
		//! Initiated by talkToGestalt..
		void talkToGestalt(Gestalt &otherG, bool checkBonds, const bool isPeriodic[3], const doublevec &cellSize);
		
		//! ...and then handled by calculateImpact of the other gestalt.
		//! For isPeriodic == true, the caluclation is repeated with mirror images
		void calculateImpact(fullForce &otherForce, Gestalt *g, bool checkBonds, const bool isPeriodic[3], const doublevec &cellSize);

		//! Calculate the external forces acting on the molecule if there exists an ExternalShape.
		void calculateExternalForce();
		
		//! Calculate and apply corrections for finite damping for Langevin propagation
		//! give externalForce and randomForce to geomShape, which replaces them with the corrected values
		//! (=> velocity is hidden in geomshape)
		void correctDamping();

		//! sum up the various forces, apply final HI scaling and perform the actual move
		//! the Gestalt does not move, when the enclosing molecule was immobilized -- 
		//! then a local counter is decremented
		void makeBrownianStep();
		//! @}

		
		//! @label Calculating HI
		//! @{
		//! Initialize the forces in the HiShape, returns number of HiShapes.
		//! Does nothing if no HiShape present.
		long int initHiForces();
		//! Sums up the effective forces, returns sum(Dij/Dii) for this Gestalt-gestalt-pair
		double talkHi(Gestalt &otherGest, const bool isPeriodic[3], const doublevec &cellSize);
		//! Calculate the effective HI-corrected sum of forces. 
		//! Just adds external and random forces when no HiShape present
		void doEffectiveForces(const double &beta_prime);
		//! @}
		
		
		//! @name immobilizing the Gestalt
		//! @{
		//! immobilize the gestalt for n timesteps
		void immobilize(long int n) {
			immobilizeCounter = n;
		}
		//! return status of the counter, asks first gestalt
		long int getImmobilized() {
			return( immobilizeCounter);
		}
		// @}

		
		// private attributes
		private:
		GeomShape  *myGeomShape;
		EstatShape *myEstatShape;
		vdWShape *myvdWShape;
		HBondShape *myhbondShape;
		BondShape *mybondShape;
		HiShape *myHiShape;
		ExternalShape *myExternalShape;
		MarkerShape *myMarkerShape;
		position *myWallsPosition;
		direction *myWallsNormal;
		
		std::string myLabel;
		
		fullPosition myPosition;
		fullPosition myActualPosition;
		
		//! @name volatile_attributes : Members which are altered throughout the lifespan of the molecule
		//@{
		//! an intermediate value: the actual force that acts on this molecule.
		//! it is calculated in one cup loop and used when called move() in the next loop
		fullForce totalForce;
		fullForce randomForce;		// my own contribution to the thermal forces
		fullForce externalForce;	// all direct external forces acting on me
		fullForce ext_Correction_Force;	
		fullForce rand_Correction_Force;
		doublevec correction_Factor_Trans;
		doublevec correction_Factor_Rot;
		//@}

		//! counter for immobilizing the molecule: if >0 the molecule does 
		//! not move during makeBrownianStep() but decrements the counter
		long int immobilizeCounter;
	};
}
#endif
