// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef MOLECULE_H
#define MOLECULE_H

#include <iostream>
#include <stdio.h>

#include "gestalt.h"
#include "interactingBody.h"

#include "types.h"
#include "vectormath.h"

/**The primary object of interest here:
Contains the Gestalt and can move
 *@author jni
 */
namespace brown{
	
	// forward declaration, because wall tries to include molecule.h
	class Wall;
	
	//! @author jni, tg
	
	//! Molecule is the central class in the cup: Molecule
	//! is a container for gestalt objects, which resemble the physical parts
	//! of a molecule.
	//! The molecule also contains the list of bonds between the gestalt parts.
	class Molecule :public InteractingBody {
		// friend class Cup;
		
		
		public:
		//! Constructor to set up new molecules
		Molecule(string name, long int num);
		//! Copy-constructor with num: used to copy from prototype molecule and set
		//! the id.mnumber to num
		Molecule(const Molecule &m, long int num);
		//! Destructor
		virtual ~Molecule();
		
		
		// C++ requirement: repetition of inherited method
		void teleport(const fullPosition &p);
		
		
		//! Output: write moleculeID and full position to file fp
		void positionToSnapshot(FILE *fp);
		
		//! @name time_tracking : keep track of where and when a molecule was inserted into the cup
		//! @{
		void setInsertionTime(timestamp &t) {
			insertionTime.when = t;
		}
		void setInsertionPlace(string &label) {
			insertionTime.where = label;
		}
		whereWhen getInsertionTime() {
			return(insertionTime);
		}
		void setTakeoutTime(timestamp &t) {
			takeoutTime.when = t;
		}
		void setTakeoutPlace(string &label) {
			takeoutTime.where = label;
		}
		whereWhen getTakeoutTime() {
			return(takeoutTime);
		}
		//! @}
		
		
		//! @name Setting and getting effective A0, which is needed by the moleculeinjector for putting the complete molecule into the simulation with the correct distribution.
		//! @{
		void setA0Trans(double a) { A0Trans = a; }
		const double getA0Trans() { return(A0Trans); }
		//! @}
		
		//! @name immobilizing the molecule
		//! @{
		//
		//! immobilize the complete molecule for n timesteps, propagated down to the Gestalten
		void immobilize(long int n);
		//! return status of the counter, asks first gestalt
		long int getImmobilized();
		// @}
		
		//! Set HICutOff in all HiShapes of all Gestalten
		void setHICutOff(double c, double s);
		
		
		//! Called from loop, check for external forces, calculation is handed down to the gestalt
		void whatIsPushing();

		
		private:
		//! Average displacement of the CM of the molecule during one Brownian timestep.
		//! This value is needed for the insertion of molecules.
		//! For a bead-spring molecule it has to be set to the value observed in a simulation.
		double A0Trans;
		
		//! time when the molecule was inserted into the cup and name of the resp. injector.
		whereWhen insertionTime;
		//! time and place where the molecule was taken out from the simulation 
		//! (be fast to read this information before this doomed molecules is finally gone)
		whereWhen takeoutTime;
		
		//! vector for the bond parameters, local to the molecule
		std::vector<bondParameter> bondVector;
	};
}
#endif
