// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef TYPES_H
#define TYPES_H


#include <vector>
#include <utility>
#include <string>
#include <assert.h>
#include <iostream>

using namespace std;

namespace brown {
	
	typedef double doublevec[3];
	typedef double quaternion[4];
	
	typedef doublevec position;
	typedef doublevec direction;
	typedef doublevec velocity;
	typedef doublevec force;
	typedef doublevec torque;
	
	/** rotation is coded as: phi = angle, n the axis around which is rotated */
	struct rotation {
		double phi;
		doublevec n;
	};
	
	typedef int timestamp;
	
	struct whereWhen {
		timestamp when;
		string where;
	};
	
	struct moleculeID{
		string mclass;
		long int mnumber;
	};
	
	struct fullPosition{
		position pos;
		rotation rot;
		bool operator==(const fullPosition &);
	};
	
	struct fullVelocity{
		velocity mov;
		velocity ang;
	};

	
	struct fullForce {
		force f;
		torque t;
		void operator+=(const fullForce &f);
	};
	
	//! Enum to allow the proteinparser to selectively ignore parts of the protein definitions
	//! (SIMULATION: ignore Markers;  ANALYSIS: consider only vdW, Markers)
	enum parserContext { SIMULATION, ANALYSIS };
	
	
	typedef double charge;
	
	//! Structure for a charge in the EstatShape
	struct chargePos {
		position pos;		// absolute position of the charge
		charge ch;			// magnitude of the charge
		double r;			// shielding radius
		position dr;		// relative position of the charge w/resp. to the molecule's center
		
		chargePos();
		chargePos(double, double, double, charge, double);
	};
	
	//! Pointer to externally supplied electrostatic interaction between charges and a wall
	struct functionPointer {
		double (*interaction)(double &distance, charge &charge);
	};


	
	//! Structure that holds a vdW-sphere with position, radius, and color
	struct vdWPos {
		position pos;		// absolute position of the vdW sphere
		double r;			// radius
		position dr;		// relative position of this sphere w/resp to the mol's center
		long int color;		// index for the vdw interaction parameter matrix
		long int id;		// the sphere's own id (currently used only for excluding interactions)
		vector<long int> blacklist;	// all the other sphere ids to be ignored
		
		public:
			vdWPos();
			// vdWPos(double, double, double, double, long int);
			vdWPos(double, double, double, double, long int, long int);		// extended by own id
			void excludeID(long int);	// add to the list of ignorable ids
			bool interacts(long int id);	// check whether id is in blacklist; if yes, return false
			vector<long int> getExcluded() { return blacklist; }
			
	};

	//! structure that holds a rectangular block with a cylindrical hole along the x-axis
	//! Has three directions (ex, ey, ez), extends to (+-lx, +-ly, +-lz), 
	//! and has a cylinder of radius r along ex (r may be 0 to denote "no hole")
	struct vdWBlock {
		position pos;
		doublevec ex, ey, ez;
		double lx, ly, lz;
		double r;
		double rGrid;
		double xPeriod, phiPeriod;
		long int color_out, color_in;		// indices for the vdw interaction parameter matrix on the surface and inside the pore
		vdWBlock();
		vdWBlock(const position &_p, const doublevec &_ex, const doublevec &_ey, const doublevec &_ez,
				 const double &_lx, const double &_ly, const double &_lz, const double &_r);
		//! Give the position and radius of a sphere and return the normal off the closest surface 
		//! of the block, the distance along that direction, and the position aong the cylinder axis.
		//! Design decision: the block does not have side wall. 
		//! Outside of the hole the force is always off the x-plane. 
		bool getDistance(doublevec &normal, double &distance, double &xpos, const position &pos, const double &r);
		// @name Return unit vectors
		// @{
		const doublevec &getEx() { return(ex); };
		const doublevec &getEy() { return(ey); };
		const doublevec &getEz() { return(ez); };
		// @}
		// @name Get parameters of the grid inside the pores
		// @{
		const double &getXPeriod()   { return(xPeriod); }
		const double &getPhiPeriod() { return(phiPeriod); }
		const double &getRGrid()     { return(rGrid); }
		// @}
	};

	
	//! Available types of interactions between vdWSpheres:
	//! (i) LJ_12_6_3 -- original Lennard-Jones with powers 12, 6, and 3;
	//! (ii) HAMAKER -- van-der-Waals between solid spheres (only works for spheres with equal radius!!!)
	//! (iii) MICROBEADS -- Hamaker plus hydrophobic coating (equal radius assumed!)
	enum VDW_INTERACTION {LJ_12_6_3, HAMAKER, MICROBEADS};
	
	//! Set of parameters defining a vdW interaction
	struct vdWParameter {
		double C_12;		// Parameter for the r^12 term
		double C_6;			// Parameter for the r^6 term
		double C_3;			// Parameter for the r^3 term
		double R_0;			// Extent of the potential
		double DR;			// distance offset
		double ASS;			// Hamaker constant sphere-sphere
		double ACC;			// Interaction strength coating-coating
		double LC;			// length of the chains
		double Rmin;		// minimal allowed distance (for numerical stability)
		double Rmax;		// cut-off, only calculate interaction within Rmax
		double Agrid;		// strength of grid potential on the wall of a cylindrical hole of a block
		enum VDW_INTERACTION intType;		// to switch between different interactions
	};
	
	
	//! Structure for a reaction sphere in GeomShape
	struct geomPos {
		position pos;		// absolute position of the vdW sphere
		double r;			// radius
		position dr;		// relative position of this sphere w/resp to the mol's center
		long int color;		// index for the vdw interaction parameter matrix
		geomPos();
		geomPos(double, double, double, double, long int);
	};
	
	//! Structure of the hook-up point for a bond in BondShape
	struct bondPos {
		position pos;		// absolute position of the hook
		position dr;		// position relative to the molecule center
		long int idx;		// index for accessing the interaction parameters
		bondPos();
	};
	
	//! Parameters for a bond
	struct bondParameter {
		double R_0;			// equlibrium distance
		double k2;			// force constant for the quadratic term
		double k4;			// force constant for the quartic term
		std::string label;
	};
	
		

	enum FORCEFIELD {NONE, LINEAR, HARMONIC, SHEAR};
	
	struct externalForce {
		enum FORCEFIELD type;
		std::string label;
		position  pos;
		direction dir;
		direction flow;
		
		externalForce();
	};
	
	
	//! Hook-up point for an external force in ExternalShape
	//! Parameters and type of the external force are set during setup
	struct externalPos {
		position pos;		// absolute position of the hook
		position dr;		// position relative to the molecule center
		double alpha;		// multiplier for force (think: charge, mass..)
		
		enum FORCEFIELD fieldtype;
		position  fieldpos;
		direction fielddir;
		direction fieldflow;

		externalPos();
		void setField(const externalForce &ef);
	};
	
	
	//! Definition of a hook-up point for a (breakable) hydrogen bond
	struct hbondPos {
		position pos;		// absolute position of the donor/acceptor atom
		direction bond;		// normalized vector pointing from donor/acceptor to the hydrogen
		position dr;		// position relative to the molecule center
		double giveOrTake;	// scaling factor: +1 for standard donor, -1 for standard acceptor
		                  	// abs value less than one for weak donor/acceptor...
	};

	
	//! Structure for a marker in the MarkerShape
	struct markerPos {
		position pos;		// absolute position of the marker
		int idx;			// index of this marker
		position dr;		// relative position of the marker w/resp. to the molecule's center
		
		markerPos();
		markerPos(double, double, double, int);
	};

	
	//! Mask function to determine whether a given protein should be accepted
	struct acceptPtr {
		bool (*mask)(const double &y, const double &z);
	};
	
	//! Mask function: returns insertion position inside valid area.
	//! Also contains the insertion area (needed to determine how many molecules
	//! enter through this mask).
	struct injectPtr {
		void (*mask)(double &y, double &z);
		double area;
	};
	
	
	struct reservoir {
		double density;
		double invVol;
	};
	
	
	//! move-and-rotate matrix
	//! this is a cumbersome implementation, but hopefully numerically faster
	//! Remember: we always have to deal with 3x3 matrices in this context
	struct tripleMatrix {
		doublevec mat [3];
		inline doublevec& operator[] (unsigned int a)
		{
			return(mat[a]);
		}
		inline const doublevec& operator[] (unsigned int a) const
		{
			return(mat[a]);
		}
	};

	struct diffusionMatrix {
		struct row {
			double coeff[6];
			inline double& operator[] (int a) {return(coeff[a]);}
			inline const double& operator[] (int a) const {return(coeff[a]);}
			inline void operator= (const row& a) { for (int i=0; i<6; ++i) coeff[i] = a[i]; }
			inline void operator= (double a) {for (int i=0; i<6; ++i) coeff[i] = a;}
			};

		row mat[6];
		inline row &operator[] (int a) {return(mat[a]);}
		inline const row &operator[] (int a) const {return(mat[a]);}

		//! you can set the 4 submatrices with this function. o references the subpart.
		//! o:   0|1
		//!      ---
		//!      2|3
		inline void set(tripleMatrix & a,unsigned int o)
		{
			assert(o<=4);
			for (int i=0;i<3;++i)
			{
				for (int j=0;j<3;++j)
				{
					(*this)[3*(o/2)+i][3*(o%2)+j] = a[i][j];
				}
			}
		}

		inline void operator= (float a) {for (int i=0; i<6; ++i) mat[i] = a;}
		inline void operator= (const diffusionMatrix& src) {for (int i=0; i<6; ++i) mat[i] = src[i];}

	};
	
}
#endif
