// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "externalshape.h"

namespace brown{
	
	ExternalShape::ExternalShape() {
	}
	
	ExternalShape::ExternalShape(const ExternalShape &other){
		hooks = other.hooks;
		actual_hooks = other.actual_hooks;
	}
	
	ExternalShape::~ExternalShape(){
		hooks.clear();
		actual_hooks.clear();
	}
	
	
	std::string ExternalShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "externalshape actual positions";
		for(vector<externalPos>::iterator mlc = actual_hooks.begin(); mlc != actual_hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2]<<"  "<<(*mlc).dr[0]<<" "<<(*mlc).dr[1]<<" "<<(*mlc).dr[2] ; 
		}
		return(ausgabe.str());
	}
	

	void ExternalShape::recalculate(const fullPosition &p) 
	{
		vector<externalPos>::iterator act = actual_hooks.begin();
		for(vector<externalPos>::iterator mlc = hooks.begin(); mlc != hooks.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
		}
	}

	
	void ExternalShape::addHook(externalPos cp, const position &moleculeCenter)
	{
		if(cp.fieldtype != NONE) {
			doublevec_set(cp.dr, cp.pos);
			doublevec_sub(cp.dr, moleculeCenter);
			hooks.push_back(cp);
			actual_hooks.push_back(cp);
		} else {
			cerr << "ExternalShape::addHook: interaction is not defined yet" << endl;
			exit(-1);
		}
	}
	
	
	void ExternalShape::calculateImpact(fullForce &myForce)
	{
		force df;
		torque dt;
		double hilf;
		
		// loop over all my hooks
		for(vector<externalPos>::iterator mlc = actual_hooks.begin(); mlc != actual_hooks.end(); ++mlc)
		{
			// std::cerr  <<"\n*** ExternalShape::calculateImpact  mlc: "<< "   " << (*mlc).pos[0] << "  " << (*mlc).pos[1] << "  " << (*mlc).pos[2] << "  " <<  endl;
			
			// get force vector on hook according to external function
			switch((*mlc).fieldtype) {
				case LINEAR:
				doublevec_set(df, (*mlc).fielddir);
				break;
				
				case HARMONIC:
				doublevec_set(df, (*mlc).fieldpos);
				doublevec_sub(df, (*mlc).pos);		// vector pointing from pp to potential min. = direction of force
				df[0] *= (*mlc).fielddir[0]; 
				df[1] *= (*mlc).fielddir[1]; 
				df[2] *= (*mlc).fielddir[2]; 
				break;
				
				case SHEAR:
				doublevec_set(df, (*mlc).pos);
				doublevec_sub(df, (*mlc).fieldpos);
				hilf = doublevec_scalprod(df, (*mlc).fielddir);
				// std::cerr  <<"    ExternalShape::calculateImpact   dr: "<< "   " << df[0] << "  " << df[1] << "  " << df[2] << "   hilf = " << hilf << endl;
				doublevec_set(df, (*mlc).fieldflow);
				// std::cerr  <<"    ExternalShape::calculateImpact flow: "<< "   " << df[0] << "  " << df[1] << "  " << df[2] << endl;
				doublevec_mult(df, hilf);
				// std::cerr  <<"    ExternalShape::calculateImpact   df: "<< "   " << df[0] << "  " << df[1] << "  " << df[2] << "  alpha = " << (*mlc).alpha << endl;
				break;
				
				case NONE:
				doublevec_set(df, 0.0);
				break;
				
				default:
				std::cerr << "externalForce::interaction: unknown force type" << std::endl;
				exit(-1);
			}
			
			// multiply with weight
			doublevec_mult(df, (*mlc).alpha);

							   
			// add force to previous result
			doublevec_add(myForce.f, df);
			
			// calculate and add torque
			doublevec_kreuzprod(dt, (*mlc).dr, df);
			doublevec_add(myForce.t, dt);
			
			// std::cerr  <<"*** ExternalShape::calculateImpact   df: "<< "   " << df[0] << "  " << df[1] << "  " << df[2] << endl;
			// std::cerr  <<"*** ExternalShape::calculateImpact   dt: "<< "   " << dt[0] << "  " << dt[1] << "  " << dt[2] << endl;

		}
	}
	
			
	
	int ExternalShape::size() {
		return(hooks.size());
	}
	
	
}
