// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "interactingBody.h"

namespace brown{
	
	InteractingBody::InteractingBody(string name, long int num)
	{
		id.mclass = name;
		id.mnumber = num;
		
		doublevec_set(myPosition.pos, 0.0);
		rotation_set(myPosition.rot, 0.0, 1.0, 0.0, 0.0);		// i.e. no rotation, but performed around the x axis
	}
	
	
	// copy constructor
	InteractingBody::InteractingBody(const InteractingBody &other)
	{
		doublevec_set(myPosition.pos, other.myPosition.pos);
		rotation_set(myPosition.rot, other.myPosition.rot);
		
		bondVector = other.bondVector;
		
		// check for Gestalten vector and copy, if existing
		Gestalten = other.Gestalten;
		std::vector<Gestalt>::iterator it;
		BondShape *bs;
		for(it = Gestalten.begin(); it != Gestalten.end(); it++) {
			if((bs = it->getBondShape())) {
				bs->setParameterVector(&bondVector);
			}
		}
	}

	
	InteractingBody::~InteractingBody() 
	{
		// test for gestalten and tear them down
		Gestalten.clear();
	}
	
	
	Gestalt *InteractingBody::getGestalt(std::string label)
	{
		std::vector<Gestalt>::iterator it;
		for(it = Gestalten.begin(); it != Gestalten.end(); it++) {
			if((it->getLabel()) == label) {
				return(&(*it));
			}
		}
	}

	
	void InteractingBody::teleport(const fullPosition &p)
	{
		// set own position
		myPosition = p;	
		// and spread the word about the latest location
		for(std::vector<Gestalt>::iterator it = Gestalten.begin(); it != Gestalten.end(); it++) {
			it->teleport(p);
		}
	}
	
	
	
	void InteractingBody::move(const position &dx, const rotation &dphi){
		quaternion myQ, dQ, ergQ;
		
		// translation
		doublevec_add(myPosition.pos, dx);
		
		// rotation part (confused by all the twists and turns?)
		rotation2quaternion(myQ, myPosition.rot);
		rotation2quaternion(dQ, dphi);
		quaternion_mult(ergQ, dQ, myQ);
		quaternion2rotation(myPosition.rot, ergQ);
		
		// hey guys, time to move on
		for(std::vector<Gestalt>::iterator it = Gestalten.begin(); it != Gestalten.end(); it++) {
			it->recalculate(myPosition);
		}
	}
	
	
	const position &InteractingBody::getCenterOfMass()
	{
		std::vector<Gestalt>::iterator it;
		vdWShape *vs;
		
		double mass = 0.0;
		
		doublevec_set(myCM, 0.0);
		
		for(it=Gestalten.begin(); it != Gestalten.end(); it++) {
			it->getvdWShape()->addMyCM(myCM, mass);		// each vdWShape knows its "weight" (=volume) and absolute position
		}
		
		if(mass > 0.0) {
			doublevec_mult(myCM, 1.0/mass);
			return(myCM);
		} else {
			return(myPosition.pos);
		}
	}
	


	long int InteractingBody::addBondParameter(const double &R_0, const double &k2, const double &k4, const std::string &label)
	{
		bondParameter bp;
		
		bp.R_0   = R_0;
		bp.k2     = k2;
		bp.k4     = k4;
		bp.label = label;
		
		bondVector.push_back(bp);
		
		return(bondVector.size() - 1);
	}
	
	// as often, information just gets handed down level by level
	void InteractingBody::setvdWParameterArray(vdWParameter **Vdw_Array) {
		for(std::vector<Gestalt>::iterator it = Gestalten.begin(); it != Gestalten.end(); it++) {
			if(it->getvdWShape()) {
				it->getvdWShape()->setParameterArray(Vdw_Array);
			}
		}
	}
	
	
	void InteractingBody::dumpStructure() 
	{
		std::cout << "Protein " << id.mclass << std::endl;
		std::cout << "\tposition "<<myPosition.pos[0]<<" "<<myPosition.pos[1]<<" "<<myPosition.pos[2]<<"   "<<myPosition.rot.phi<<" "<<myPosition.rot.n[0]<<" "<<myPosition.rot.n[1]<<" "<<myPosition.rot.n[2]<<std::endl;
		
		// Dump Bonds
		if(bondVector.size() > 0) {
			std::cout << "\n\tBonds myBonds" << std::endl;
			for(std::vector<bondParameter>::iterator it=bondVector.begin(); it != bondVector.end(); it++) {
				std::cout << "\t\t" << (*it).label << "  " << (*it).R_0 << "  " << (*it).k2 << "  " << (*it).k4 << std::endl;
			}
			std::cout << "\tEnd" << std::endl;
		}
		
		// let the Gestalten speak for themselves
		for(std::vector<Gestalt>::iterator it = Gestalten.begin(); it != Gestalten.end(); it++) {
			it->dumpStructure();
		}
		
		std::cout << "End\n" << std::endl;	// "End of Story"
	}

}
