// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#include "moleculeinjector.h"


namespace brown {

	gsl_rng *MoleculeInjector::rng = gsl_rng_alloc(gsl_rng_taus);

	// constructor from scratch
	MoleculeInjector::MoleculeInjector(string newLabel) : MoleculeHandler(newLabel)
	{
		//! @todo initialize mask with default rectangular mask
		// set mask to NULL -> crash, if used before defined
		// myMask.mask = NULL;
		// myMask.area = 1.0;
		maskY0 = 0.0;
		maskZ0 = 0.0;
		maskdY = 0.0;
		maskdZ = 0.0;
		maskArea = 0.0;
		
		molTemplates.clear();
		gsl_rng_set(rng,0);
	}
	
	// copy constructor
	MoleculeInjector::MoleculeInjector(MoleculeInjector &other) : MoleculeHandler(other)
	{
		// myMask.mask = other.myMask.mask;
		// myMask.area = other.myMask.area;
		maskY0 = other.maskY0;
		maskZ0 = other.maskZ0;
		maskdY = other.maskdY;
		maskdZ = other.maskdZ;
		maskArea = other.maskArea;
		
		molTemplates = other.molTemplates;
	}
	
	
	std::string MoleculeInjector::describe() {
		ostringstream ausgabe("", ios::out);
		
		ausgabe << label << ": ";
		
		vector<insertTemplate>::iterator it;
		for(it = molTemplates.begin(); it != molTemplates.end(); ++it) {
			ausgabe << "  " << it->molName << "@" << it->resName << "(" << *(it->concentration) << ")" ;
		}
		
		return(ausgabe.str());
	}
	


	void MoleculeInjector::addMolecule(Molecule &oldMol, double *concentration, const double &invVol, vector<fullPosition> *insPos, int *theInternal, int resInc, long int immobilizeT)
	{
		cout << "### MoleculeInjector::addMolecule: long version" << endl;
		
		insertTemplate newT;
		
		newT.mol = new Molecule(oldMol, 0L);
		newT.mol->setInsertionPlace(label);
		newT.molName = oldMol.getid().mclass;
		newT.resName = "";
		newT.concentration = concentration;
		newT.invVol = invVol;
		newT.insPos = insPos;
		newT.theInternal = theInternal;
		newT.resInc = resInc;
		newT.immobilizeT = immobilizeT;
		
		molTemplates.push_back(newT);
	}
	

	void MoleculeInjector::addMolecule(std::string molName, std::string resName)
	{
		cout << "### MoleculeInjector::addMolecule: short version" << endl;
		
		insertTemplate newT;
		
		newT.mol = NULL;
		// newT.mol->setInsertionPlace(label);
		newT.molName = molName;
		newT.resName = resName;
		newT.concentration = NULL;
		newT.invVol = 0.0;
		newT.insPos = NULL;
		newT.theInternal = NULL;
		newT.resInc = 0;
		newT.immobilizeT = 0L;

		molTemplates.push_back(newT);
		
		std::cerr << "MoleculeInjector::addMolecule: added " << molTemplates[molTemplates.size()-1].molName << " - " <<  molTemplates[molTemplates.size()-1].resName<< endl;
		// std::cerr << "MoleculeInjector::addMolecule: size(molTemplates) = " << molTemplates.size() << endl;
		// for(vector<insertTemplate>::iterator it = molTemplates.begin(); it != molTemplates.end(); ++it) {
		// 	std::cerr << "MoleculeInjector::addMolecule: listing: " << it->molName << ", " << it->resName << std::endl;
		// }
	}
	
	
	
	void MoleculeInjector::resolveMolNames(std::map<std::string, Molecule*> *molMap) 
	{
		vector<insertTemplate>::iterator it;
		for(it = molTemplates.begin(); it != molTemplates.end(); ++it)
		{
			if(molMap->count(it->molName) == 1) {
				it->mol = (*molMap)[it->molName];
				std::cerr << "MoleculeInjector::resolveMolNames: resolved molecule "<< it->molName  << " for reservoir " << it->resName << std::endl;
			}
			else {
				std::cerr << "MoleculeInjector::resolveMolNames: cannot resolve molecule " << it->molName << std::endl;
				exit(-1);
			}
		}
	}
	
	
	
	void MoleculeInjector::resolveResNames(std::map<std::string, struct reservoir> *resMap) 
	{
		vector<insertTemplate>::iterator it;
		for(it = molTemplates.begin(); it != molTemplates.end(); ++it)
		{
			if(resMap->count(it->resName) == 1) {
				it->concentration = &((*resMap)[it->resName].density);
				it->invVol = (*resMap)[it->resName].invVol;
				std::cerr << "MoleculeInjector::resolveResNames: resolved reservoir " << it->resName << " with concentration = " << *(it->concentration) << std::endl;
			}
			else {
				std::cerr << "MoleculeInjector::resolveResNames: cannot resolve reservoir " << it->resName << std::endl;
				exit(-1);
			}
		}
	}
	
	
	void MoleculeInjector::insert()
	{
		const double insertDepth = 4.5;		// the probability that a molecule is inserted beyond 4.5 A0 is negligible
		double dV, c, iProb, pTest, a0;
		double xTest, xInsert, yInsert, zInsert;
		double phiTest, thetaTest;
		int nNew;
		position helpPos;
		fullPosition insPos;
		
		int gesamtZahl = 0;
		
		// cout << "MoleculeInjector::insert: num of templates = " << molTemplates.size() << endl;
		// cout << "    position = " << (*myWallsPosition)[0] << " " << (*myWallsPosition)[1] << " " << (*myWallsPosition)[2] << endl;
		// cout << "    normal   = " << (*myWallsNormal)[0] << " " << (*myWallsNormal)[1] << " " << (*myWallsNormal)[2] <<endl;
		// cout << "    y axis   = " << (*myWallsYAxis)[0] << " " << (*myWallsYAxis)[1] << " " << (*myWallsYAxis)[2] <<endl;
		// cout << "    z axis   = " << (*myWallsZAxis)[0] << " " << (*myWallsZAxis)[1] << " " << (*myWallsZAxis)[2] <<endl;
		
		// loop over all mol templates
		int i=1;
		for(vector<insertTemplate>::iterator it = molTemplates.begin(); it != molTemplates.end(); ++it, ++i)
		{
			moleculeID mid = (*it).mol->getid();
			
			// std::cerr << "MoleculeInjector::insert: template nr. " << i << endl;
			
			a0 = (*it).mol->getA0Trans();
			// std::cerr << "### MoleculeInjector::insert: A0    = " << a0 << endl;
			
			// std::cerr << "MoleculeInjector::inject: anfangs invVol = " << (*it).invVol << "  density = " <<(*((*it).concentration)) << std::endl;

			dV = maskArea * insertDepth * a0;
			c = (*((*it).concentration));
			// if negative vol (number counter) then nNew = conc.
			if((*it).invVol >= 0.0) {
				nNew = int(0.5* dV * c + 1.0);
				// iProb = 0.5 * dV * c / nNew;
				iProb = 0.5 * dV * c / double(nNew);
			} else {
				nNew = int((*((*it).concentration)));
// 				if(nNew>0) {
// 						cout << "### MoleculeInjector::insert: inserting " << nNew << " molecules " << mid.mclass << " from number res." << endl;
// 						if((*it).insPos) {
// 							cout << "### MoleculeInjector::insert: length of pos vector = " << (*(*it).insPos).size() << endl;
// 						}
// 				}
			}
			
			
			// cout << "### MoleculeInjector::insert: dV    = " << dV << endl;
			// cout << "MoleculeInjector::insert: nNew  = " << nNew << endl;
			// std::cerr << "MoleculeInjector::insert: iProb = " << iProb << endl;
			// std::cerr << "MoleculeInjector::inject: A invVol = " << (*it).invVol << "  density = " <<(*((*it).concentration)) << std::endl;

			// now try for nNew times to insert a molecule
			for(int k=0; k<nNew; ++k)
			{
				xTest = insertDepth * ran2();
				
				// check for negative vol -> concentration is a particle counter
				if((*it).invVol >= 0.0) {
					// xTest = insertDepth * ran2();
					pTest = iProb * erfcc(xTest);
				} else {
					pTest = 1.1;
				}
				
				// shall we choose this one?
				if(pTest >= ran2())
				{
					++gesamtZahl;
					// decrease concentration of remaining molecules
					if((*it).invVol >= 0.0) {
						(*((*it).concentration)) -= (*it).invVol;
					} else {
						// cout << "   before reducing density, d = " << (*((*it).concentration)) << endl;
						(*((*it).concentration)) -= 1.0;
						// cout << "   after reducing density" << endl;
					}
					
					// cout << "### MoleculeInjector:: injected molecule" << endl;
					
					// check if saved insertion positions are available, if so, use these
					if((*it).insPos) 
					{
						insPos = (*(*it).insPos)[k];
						// cout << "### MoleculeInjector::insert: Injector "<<label<<",  molecule " <<mid.mclass<< " at " <<insPos.pos[0]<<" "<<insPos.pos[1]<<" "<<insPos.pos[2]<< " ("<< insPos.pos <<")"<<endl;
						// cout << "### MoleculeInjector::insert: Injector "<<label<<",  molecule " <<mid.mclass<< " at " <<insPos.pos[0]<<" "<<insPos.pos[1]<<" "<<insPos.pos[2]<<endl;
					} 
					else 
					{
						xInsert = a0 * xTest;
						
						// myMask.mask(yInsert, zInsert);
						//! @todo Replace hard coded rectangular mask with mask function (default should be rectangular with dimensions form moleculehandler)
						yInsert = (1.0 - 2.0 * ran2()) * maskdY + maskY0;
						zInsert = (1.0 - 2.0 * ran2()) * maskdZ + maskZ0;
						
						// if((*it).invVol < 0.0) {
						// 	cout << "   after mask" << endl;
						// std::cerr << "### MoleculeInjector:: inject molecule at " << xInsert <<" "<<yInsert<<" "<<zInsert<<endl;
						// }
						
						// std::cerr << "Moleculeinjector::insert: myWallsNormal = " <<(*myWallsNormal)[0]<<" "<<(*myWallsNormal)[1]<<" "<<(*myWallsNormal)[2]<< std::endl;
						// std::cerr << "Moleculeinjector::insert: myWallsYAxis  = " <<(*myWallsYAxis)[0]<<" "<<(*myWallsYAxis)[1]<<" "<<(*myWallsYAxis)[2]<< std::endl;
						// std::cerr << "Moleculeinjector::insert: myWallsZAxis  = " <<(*myWallsZAxis)[0]<<" "<<(*myWallsZAxis)[1]<<" "<<(*myWallsZAxis)[2]<< std::endl;
						
						// calculate insertion position of molecule in cup's frame
						doublevec_set(insPos.pos, (*myWallsPosition));
						
						doublevec_set(helpPos, (*myWallsNormal));
						doublevec_mult(helpPos, xInsert);
						doublevec_add(insPos.pos, helpPos);
						
						doublevec_set(helpPos, (*myWallsYAxis));
						doublevec_mult(helpPos, yInsert);
						doublevec_add(insPos.pos, helpPos);
						
						doublevec_set(helpPos, (*myWallsZAxis));
						doublevec_mult(helpPos, zInsert);
						doublevec_add(insPos.pos, helpPos);
						
						// choose any orientation
						phiTest = 2.0*M_PI * ran2();
						thetaTest = acos(1.0 - 2.0* ran2());
						rotation_set(insPos.rot, 2.0*M_PI* ran2(), cos(phiTest)*sin(thetaTest), sin(phiTest)*sin(thetaTest), cos(thetaTest));
					}
					
					// move the template mol to the insertion position and put it into the cup
					(*it).mol->teleport(insPos);

					// immobilize molecule
					(*it).mol->immobilize((*it).immobilizeT);
					
					position hier;
					doublevec_set(hier, (*it).mol->getPosition());

					if(myWallsCup) {
						moleculeID eidi = (*it).mol->getid();
						myWallsCup->addMolecule(*((*it).mol));
					} else {
						cerr << "### MoleculeInjector::insert: NO CUP DEFINED ###  myWallsCup = " << myWallsCup<< endl;
					}

					// now clean up: delete insertion positions
					if((*it).insPos) {
						(*(*it).insPos).clear();
					}
					
					if((*it).theInternal) {
						// cout << "    internal reservoir: " << (*((*it).theInternal));
						(*((*it).theInternal)) += (*it).resInc;
						// cout << " -> " << (*((*it).theInternal)) << endl;
					}

					
					// if((*it).invVol < 0.0) {
					// 	cout << "    mol inserted" << endl;
					// }
				}
				 
			}
		}

// 		if(gesamtZahl > 0) {
// 			cout << "*** MoleculeInjector::insert: inserted " << gesamtZahl << " molecules" << endl;
// 			cout << "### MoleculeInjector::insert: A0    = " << a0 << endl;
// 			cout << "### MoleculeInjector::insert: dV    = " << dV << endl;
// 			cout << "### MoleculeInjector::insert: nNew  = " << nNew << endl;
// 			cout << "### MoleculeInjector::insert: iProb = " << iProb << endl;
// 		}
	}
	 
}
