// This file is part of the brownmove simulation package
// 
// V1.1  Oct 2010
// 
// (C) 2003 - 2010 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove


#include "vdwshape.h"

namespace brown{
	
	vdWShape::vdWShape(){
		paramsArray = NULL;
		myWallColor = -1;
		myWallOffset = 0.0;
		doublevec_set(myActualWallPos, 0.0);
		doublevec_set(myActualWallNormal, 1.0, 0.0, 0.0);
	}
	
	vdWShape::vdWShape(const vdWShape &other){
		spheres = other.spheres;
		actual_spheres = other.actual_spheres;
		blocks = other.blocks;
		actual_blocks = other.actual_blocks;
		paramsArray = other.paramsArray;
		myWallColor = other.myWallColor;
		myWallOffset = other.myWallOffset;
		doublevec_set(myActualWallPos, other.myActualWallPos);
		doublevec_set(myActualWallNormal, other.myActualWallNormal);
	}
	
	vdWShape::~vdWShape(){
		spheres.clear();
		actual_spheres.clear();
		blocks.clear();
		actual_blocks.clear();
	}
	
	
	std::string vdWShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual sphere positions";
		vector<vdWPos>::iterator mlc;
		for(mlc = actual_spheres.begin(); mlc != actual_spheres.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		// ausgabe <<" :";
		// for(mlc = spheres.begin(); mlc != spheres.end(); ++mlc, i++){
		// 	ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		// }
		
		if(actual_blocks.size() > 0) {
			ausgabe << "  blocks";
			vector<vdWBlock>::iterator ibl;
			for(ibl = actual_blocks.begin(); ibl != actual_blocks.end(); ibl++) {
				ausgabe << " : " << (*ibl).pos[0]<<" "<<(*ibl).pos[1]<<" "<<(*ibl).pos[2];
			}
		}
				
		if(myWallColor >= 0) {
			ausgabe << "  wall: " <<myActualWallPos[0]<<" "<<myActualWallPos[1]<<" "<<myActualWallPos[2]<<" / "<<myActualWallNormal[0]<<" "<<myActualWallNormal[1]<<" "<<myActualWallNormal[2];
		}
		
		return(ausgabe.str());
	}
	
	
	
	void vdWShape::recalculate(const fullPosition &p) 
	{
		vector<vdWPos>::iterator act = actual_spheres.begin();
		for(vector<vdWPos>::iterator mlc = spheres.begin(); mlc != spheres.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			doublevec_set((*act).dr, (*act).pos);
			positionMove((*act).pos, p.pos);
		}

		if(blocks.size() > 0) 
		{
			vector<vdWBlock>::iterator abl = actual_blocks.begin();
			for(vector<vdWBlock>::iterator ibl = blocks.begin(); ibl != blocks.end(); ++ibl, ++abl) 
			{
				// position
				doublevec_set((*abl).pos, (*ibl).pos);
				doublevec_rotate((*abl).pos, p.rot);
				positionMove((*abl).pos, p.pos);
				
				// coordinate system - only rotation
				doublevec_set((*abl).ex, (*ibl).ex);
				doublevec_rotate((*abl).ex, p.rot);
				doublevec_set((*abl).ey, (*ibl).ey);
				doublevec_rotate((*abl).ey, p.rot);
				doublevec_set((*abl).ey, (*ibl).ey);
				doublevec_rotate((*abl).ey, p.rot);
			}
		}
		
		if(myWallColor >= 0) 
		{
			doublevec shift;
			doublevec_set(shift, 1.0, 0.0, 0.0);		// this was the original normal of the wall in the un-teleported gestalt
			doublevec_rotate(shift, p.rot);				// new direction of the normal vector
			doublevec_set(myActualWallNormal, shift);	// save normal
			doublevec_mult(shift, myWallOffset);		// shift distance
			
			doublevec_set(myActualWallPos, p.pos);	// new position of the wall
			doublevec_add(myActualWallPos, shift);	// shift along normal
		}
	}

	

	void vdWShape::addSphere(vdWPos vp, const position &moleculeCenter){
		doublevec_set(vp.dr, vp.pos);
		doublevec_sub(vp.dr, moleculeCenter);
		spheres.push_back(vp);
		actual_spheres.push_back(vp);
	}

	
	
	void vdWShape::addBlock(vdWBlock vb) {
		blocks.push_back(vb);
		actual_blocks.push_back(vb);
	}
	
	
	double vdWShape::interaction(double &dist, const long int &clr1, const long int &clr2, const double &r1, const double &r2)
	{
		
		double prefactor = 0.0;
		double r12 = dist + r1 + r2; 
		
		// std::cerr << "vdWShape::interaction:  dist = " << dist << "  r12 = " << r12 << "  type = " << paramsArray[clr1][clr2].intType << std::endl;
		
		switch(paramsArray[clr1][clr2].intType) {
			// ---- Lennard-Jones with powers 12, 6, 3 (the original version) ----
			case (LJ_12_6_3):
			{
				dist += paramsArray[clr1][clr2].DR;
				
				// std::cerr << "vdWShape::interaction: LJ_12_6_3:  dist = " << dist << "  r12 = " << r12 << std::endl;
				
				double rr1 = paramsArray[clr1][clr2].R_0 / dist;		// powers of the inverse distance
				double rr3 = rr1 * rr1 * rr1;
				double rr4 = rr3 * rr1;
				double rr6 = rr3 * rr3;
				
				prefactor = -((12.0*paramsArray[clr1][clr2].C_12 * rr6 + 6.0*paramsArray[clr1][clr2].C_6) * rr3 + 3.0*paramsArray[clr1][clr2].C_3) * rr4 / paramsArray[clr1][clr2].R_0;
			}
			break;
			
			// ---- Repulsive interaction from coating (Hamaker style vdW attraction is added below) ----
			case (MICROBEADS):
			{
				// std::cerr << "vdWShape::interaction: MICROBEADS:  dist = " << dist << "  r12 = " << r12 << std::endl;
				
				double RL = r1 + paramsArray[clr1][clr2].LC;
				if(r12 <= 2.0 * RL)
				{
					double RL2 = RL*RL;
					double Rtilde = r1 - paramsArray[clr1][clr2].DR;
					
					double nullH = Rtilde /(r12 - RL);
					double zweiH = 1 - nullH*nullH;
					double einsH = 2.0*Rtilde/RL2 * nullH*nullH / (zweiH*zweiH) * (2.0/3.0*RL2*RL - 0.5*RL2*r12 + r12*r12*r12/24.0);
					
					double dreiH = Rtilde /RL;
					double vierH = dreiH * dreiH / zweiH;
					vierH *= (0.5 * RL2 - 0.125*r12*r12);
					
					prefactor = -paramsArray[clr1][clr2].ACC * (einsH + vierH);
					
					// std::cerr << "vdWShape::interaction: dist = " << dist << "  r12 = " << r12 << "  prefactor = " << prefactor << std::endl;
				} else {
					
					prefactor = 0.0;
				}
			}
			
			// ---- Hamaker style van-der-Waals interaction ----
			case (HAMAKER):
			{
				double r12_2 = r12*r12;
				double r0_2 = r1*r1;
				double r4r = r12_2 - 4.0*r0_2;
				
				double einsL = 4.0*r0_2*r12 / (r4r*r4r);
				double zweiL = 4*r0_2 / (r12*r12_2);
				double dreiL = 8.0*r0_2 / (r12 * r4r);
				
				prefactor += paramsArray[clr1][clr2].ASS/6.0 * (einsL + zweiL - dreiL);	// relies upon the initialization of prefactor = 0.0 for "only Hamaker"!!!
			}
			break;
			
			// we should never get here (as long as the parser is used)
			default:
			std::cerr << "vdWShape::interaction: undefined interaction type - aborting.." << std::endl;
			exit(-1);
		}
		
		// if everything worked well: return calculated force
		return(prefactor);
	}
	

	
	void vdWShape::calculateImpact(fullForce &myForce, fullForce &otherForce, vdWShape *vs, const bool isPeriodic[3], const doublevec &cellSize)
	{
		position dr, ri, dr_temp;
		double prefactor, B12, R0;
		double r12, dist, rr1, rr3, rr4, rr6; //, rr7;
		force df;
		torque dt;
		long int clr1, clr2;
		long int id2;			// andi
		
		// first interact all spheres with each other -- nothing happens, if there are none
		// Therefore, begin with the other's spheres (could be a wall..) -- we have at least one vdw sphere, or we would not be here.
		for(vector<vdWPos>::iterator mlc2 = vs->actual_spheres.begin(); mlc2 != vs->actual_spheres.end(); ++mlc2)
		{
			clr2 = (*mlc2).color;
			id2 = (*mlc2).id;		// andi
			
			for(vector<vdWPos>::iterator mlc1 = actual_spheres.begin(); mlc1 != actual_spheres.end(); ++mlc1)
			{
				if ((*mlc1).interacts(id2)) {
					// std::cerr << "vdwshape::calculateImpact: excluded ids " << (*mlc1).id << " and " << id2 <<  std::endl;	
					continue;
				}
				// std::cerr  << "vdwshape::calculateImpact: now for the interaction..." << std::endl;
				
				clr1 = (*mlc1).color;
				
				// calculate relative distances etc
				doublevec_set(dr, (*mlc2).pos);
				doublevec_sub(dr, (*mlc1).pos);
				B12 = (*mlc1).r + (*mlc2).r;
				
				double dx = 0.0;
				double dy = 0.0;
				double dz = 0.0;
				int ix = 2, iy = 2, iz = 2;
				
				if(isPeriodic[0]) {
					dx = -cellSize[0];
					ix = 0;
				}
				if(isPeriodic[1]) {
					dy = -cellSize[1];
					iy = 0;
				}
				if(isPeriodic[2]) {
					dz = -cellSize[2];
					iz = 0;
				}
				
				while(true)
				{
					dr_temp[0] = dr[0] + dx;
					dr_temp[1] = dr[1] + dy;
					dr_temp[2] = dr[2] + dz;
					
					r12 = doublevec_length(dr_temp);
					
					dist = r12 - B12;		// empty space between the spheres <- determines the interaction
					
					// check whether dist > cut-off  =>  skip further caluclation
					if((dist <= paramsArray[clr1][clr2].Rmax))
					{
						
						// check for minimal distance for numerical stability
						if(dist < paramsArray[clr1][clr2].Rmin) {
							// std::cerr << "vdWShape::calculateImpact: dist = " << dist << " < " << paramsArray[clr1][clr2].Rmin <<  std::endl;
							dist = paramsArray[clr1][clr2].Rmin;
						}
						
						// --- here comes the interaction itself
						prefactor = interaction(dist, clr1, clr2, (*mlc1).r, (*mlc2).r);
						
						doublevec_set(df, dr_temp);					// for the direction
						doublevec_mult(df, (prefactor/r12));	// normalize dr_temp by its length
						
						// add force to previous results
						doublevec_add(myForce.f, df);
						doublevec_sub(otherForce.f, df);
						
						// calculate and add torques
						doublevec_kreuzprod(dt, (*mlc1).dr, df);
						doublevec_add(myForce.t, dt);
						doublevec_kreuzprod(dt, (*mlc2).dr, df);
						doublevec_sub(otherForce.t, dt);
					}
					
					// this part does the periodic boundary conditions
					if(ix < 2) {
						ix++;
						dx += cellSize[0];
					}
					else
					{
						if(isPeriodic[0]) {
							ix = 0;
							dx = -cellSize[0];
						}
						
						if(iy < 2) {
							iy++;
							dy += cellSize[1];
						}
						else
						{
							if(isPeriodic[1]) {
								iy = 0;
								dy = -cellSize[1];
							}
							
							if(iz < 2) {
								iz++;
								dz += cellSize[2];
							}
							else
							{
								break;
							}
						}
					}
				}
			}
		}
		
		
		// calculate interaction with vdw blocks
		// As blocks do not have an interaction off the side walls, there is no reason to apply periodic boundary conditions
		// => blocks are supposed to span the complete simulation volume from wall to wall (use multiple blocks?)
		for(vector<vdWBlock>::iterator mlc2 = vs->actual_blocks.begin(); mlc2 != vs->actual_blocks.end(); ++mlc2)
		{
			doublevec distV, fx, fphi;
			double abstand, xpos;
			double dummy = 0.0;
			
			for(vector<vdWPos>::iterator mlc1 = actual_spheres.begin(); mlc1 != actual_spheres.end(); ++mlc1)
			{
				clr1 = (*mlc1).color;
				
				// std::cerr << "vdWShape::calculateImpact: block  start, mol pos = "<<(*mlc1).pos[0]<<", "<<(*mlc1).pos[1]<<", "<<(*mlc1).pos[2] << std::endl;
				
				// calculate distance and direction, do nothing if outside of bounds
				if((*mlc2).getDistance(distV, abstand, xpos, (*mlc1).pos, (*mlc1).r))
				{
					// std::cerr << "vdWShape::calculateImpact: block  got distance" << std::endl;

					// use respective color when inside the pore (xpos > 0) or outside of pore
					if(xpos > 0.0) {
						clr2 = (*mlc2).color_in; 
					} else {
						clr2 = (*mlc2).color_out;
					}
					
					// check whether dist > cut-off  =>  skip further caluclation
					if(abstand <= paramsArray[clr1][clr2].Rmax)
					{
						// std::cerr << "vdWShape::calculateImpact: block  abstand = " << abstand << std::endl;
						
						// check for minimal distance
						if(abstand < paramsArray[clr1][clr2].Rmin) {
							abstand = paramsArray[clr1][clr2].Rmin;
						}
						
						// --- here comes the interaction itself
						// note the "-" sign -- different point of view than with the spheres
						prefactor = -interaction(abstand, clr1, clr2, (*mlc1).r, dummy);

						// calculate the force normal to the wall (normal is normalized -- should be :-))
						doublevec_set(df, distV);
						doublevec_mult(df, prefactor);
						
						// std::cerr << "                               df = " << df[0]<<", "<<df[1]<<", "<<df[2] << std::endl;
						
						// determine force along ex and along the circumference if strength > 0
						if(xpos > 0.0) {
							// std::cerr << "vdWShape::calculateImpact: xpos = "<<xpos<<std::endl;
							
							if(paramsArray[clr1][clr2].Agrid > 0.0)
							{
								// decay with distance off the cylinder wall
								double help = abstand / (*mlc2).getRGrid();
								double prefR = exp(-0.5*help*help) * paramsArray[clr1][clr2].Agrid;
								
								// periodicity along axis
								double k = 2.0*M_PI / (*mlc2).getXPeriod();
								double prefX = k * sin(k * xpos);
								doublevec_set(fx, (*mlc2).getEx());
								doublevec_mult(fx, prefX*prefR);
								
								// periodicity along circumference
								double cy = doublevec_scalprod((*mlc2).getEy(), distV);
								double cz = doublevec_scalprod((*mlc2).getEz(), distV);
								double winkel = atan2(cz, cy);
								k = (*mlc2).getPhiPeriod();
								double prefW = k * sin(k * winkel);
								doublevec_kreuzprod(fphi, (*mlc2).getEx(), distV);
								doublevec_mult(fphi, prefW*prefR);
								
								// std::cerr << "vdWShape::calculateImpact: inside: xpos = "<<xpos<<"   fx = "<<fx[0]<<", "<<fx[1]<<", "<<fx[2]<<std::endl;
								
								// add to radial force
								doublevec_add(df, fx);
								doublevec_add(df, fphi);
							}
						}
						// std::cerr << "                               df = " << df[0]<<", "<<df[1]<<", "<<df[2] << std::endl;

						doublevec_add(myForce.f, df);
						doublevec_kreuzprod(dt, (*mlc1).dr, df);
						doublevec_add(myForce.t, dt);
					}
				}
			}
		}
			
			
		// now -- for walls -- check, if the global vdWColor is >= 0 (should remain unset (-1) for molecules built from vdw spheres)
		// Again, periodicity does not make much sense parallel to an infinitely extending wall with an interaction directly off it
		if((clr2 = vs->getWallColor()) >= 0)
		{
			position *wallPos;
			doublevec *wallNormal;
			position abstand;
			double dummy = 0.0;
			
			wallPos = vs->getWallPosition();
			wallNormal = vs->getWallNormal();
			
			int i = 1;
			for(vector<vdWPos>::iterator mlc1 = actual_spheres.begin(); mlc1 != actual_spheres.end(); ++mlc1, ++i)
			{
				clr1 = (*mlc1).color;
				
				doublevec_set(abstand, (*mlc1).pos);
				doublevec_sub(abstand, (*wallPos));
				
				// empty space between the mol and the wall, measured normal to the wall (= shortest distance)
				dist = doublevec_scalprod(abstand, (*wallNormal)) - (*mlc1).r;
				
				// check whether dist > cut-off  =>  skip further caluclation
				if((dist <= paramsArray[clr1][clr2].Rmax))
				{
					// check for minimal distance
					if(dist < paramsArray[clr1][clr2].Rmin) {
						dist = paramsArray[clr1][clr2].Rmin;
					}
					
 					// --- here comes the interaction itself
					// note that the "-" sign: different point of view than with the spheres
					prefactor = -interaction(dist, clr1, clr2, (*mlc1).r, dummy);

					// calculate the force normal to the wall (normal is normalized -- should be :-))
					doublevec_set(df, (*wallNormal));
					doublevec_mult(df, prefactor);
					
					doublevec_add(myForce.f, df);
					doublevec_kreuzprod(dt, (*mlc1).dr, df);
					doublevec_add(myForce.t, dt);
				}
			}
		}
	}
	
	
	int vdWShape::size() {
		return(spheres.size());
	}
	
	
	void vdWShape::addMyCM(position &theCM, double &theMass)
	{
		doublevec tmppos;
		double m;
		
		for(vector<vdWPos>::iterator it = actual_spheres.begin(); it != actual_spheres.end(); it++)
		{
			m = (*it).r;
			m *= (m*m);		// mass propto r^3

			doublevec_set(tmppos, (*it).pos);
			// std::cerr << "vdWShape::addMyCM: pos = " << tmppos[0]<<", "<<tmppos[1]<<", "<<tmppos[2]<< std::endl;
			
			doublevec_mult(tmppos, m);
			doublevec_add(theCM, tmppos);
			theMass += m;
			
			// std::cerr << "vdWShape::addMyCM: r = " << (*it).r << "  m = " << m << "   tmppos = " << tmppos[0]<<", "<<tmppos[1]<<", "<<tmppos[2]<< std::endl;
		}
	}
	
	
	void vdWShape::dumpStructure() {
		std::cout << "\n\t\tVdwShape mySpheres" << std::endl;
		
		for(std::vector<vdWPos>::iterator it = spheres.begin(); it != spheres.end(); it++) {
			std::cout << "\t\t\t"<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<< (*it).r << "   " << (*it).color << std::endl;
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	
	void vdWShape::dumpSpheres(char *prefix) 
	{
		int i = 1;
		for(std::vector<vdWPos>::iterator it = actual_spheres.begin(); it != actual_spheres.end(); it++, i++) {
			std::cout << prefix <<" " << i << " "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<" "<< (*it).r << " " << (*it).color << std::endl;
		}

	}

	
	int vdWShape::dumpSpheres(char *prefix, string *raus) 
	{
		int i = 1;
		ostringstream ausgabe("", ios::out);

		for(std::vector<vdWPos>::iterator it = actual_spheres.begin(); it != actual_spheres.end(); it++, i++) {
			ausgabe << prefix << i << " "<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<< std::endl;
		}
		
		(*raus) += ausgabe.str();
		return(i-1);
	}

}
